@@ -9,8 +9,8 @@ use ide::{
     Annotation, AnnotationKind, Assist, AssistKind, CallInfo, Cancellable, CompletionItem,
     CompletionItemKind, CompletionRelevance, Documentation, FileId, FileRange, FileSystemEdit,
     Fold, FoldKind, Highlight, HlMod, HlOperator, HlPunct, HlRange, HlTag, Indel, InlayHint,
-    InlayKind, InsertTextFormat, Markup, NavigationTarget, ReferenceAccess, RenameError, Runnable,
-    Severity, SourceChange, StructureNodeKind, SymbolKind, TextEdit, TextRange, TextSize,
+    InlayKind, Markup, NavigationTarget, ReferenceAccess, RenameError, Runnable, Severity,
+    SourceChange, StructureNodeKind, SymbolKind, TextEdit, TextRange, TextSize,
 };
 use itertools::Itertools;
 use serde_json::to_value;
@@ -92,15 +92,6 @@ pub(crate) fn documentation(documentation: Documentation) -> lsp_types::Document
     lsp_types::Documentation::MarkupContent(markup_content)
 }
 
-pub(crate) fn insert_text_format(
-    insert_text_format: InsertTextFormat,
-) -> lsp_types::InsertTextFormat {
-    match insert_text_format {
-        InsertTextFormat::Snippet => lsp_types::InsertTextFormat::Snippet,
-        InsertTextFormat::PlainText => lsp_types::InsertTextFormat::PlainText,
-    }
-}
-
 pub(crate) fn completion_item_kind(
     completion_item_kind: CompletionItemKind,
 ) -> lsp_types::CompletionItemKind {
@@ -197,36 +188,67 @@ pub(crate) fn snippet_text_edit_vec(
         .collect()
 }
 
-pub(crate) fn completion_item(
-    insert_replace_support: Option<lsp_types::Position>,
+pub(crate) fn completion_items(
+    insert_replace_support: bool,
+    enable_imports_on_the_fly: bool,
     line_index: &LineIndex,
-    item: CompletionItem,
+    tdpp: lsp_types::TextDocumentPositionParams,
+    items: Vec<CompletionItem>,
 ) -> Vec<lsp_types::CompletionItem> {
+    let max_relevance = items.iter().map(|it| it.relevance().score()).max().unwrap_or_default();
+    let mut res = Vec::with_capacity(items.len());
+    for item in items {
+        completion_item(
+            &mut res,
+            insert_replace_support,
+            enable_imports_on_the_fly,
+            line_index,
+            &tdpp,
+            max_relevance,
+            item,
+        )
+    }
+    res
+}
+
+fn completion_item(
+    acc: &mut Vec<lsp_types::CompletionItem>,
+    insert_replace_support: bool,
+    enable_imports_on_the_fly: bool,
+    line_index: &LineIndex,
+    tdpp: &lsp_types::TextDocumentPositionParams,
+    max_relevance: u32,
+    item: CompletionItem,
+) {
     let mut additional_text_edits = Vec::new();
-    let mut text_edit = None;
+
     // LSP does not allow arbitrary edits in completion, so we have to do a
     // non-trivial mapping here.
-    let source_range = item.source_range();
-    for indel in item.text_edit().iter() {
-        if indel.delete.contains_range(source_range) {
-            text_edit = Some(if indel.delete == source_range {
-                self::completion_text_edit(line_index, insert_replace_support, indel.clone())
+    let text_edit = {
+        let mut text_edit = None;
+        let source_range = item.source_range();
+        for indel in item.text_edit().iter() {
+            if indel.delete.contains_range(source_range) {
+                let insert_replace_support = insert_replace_support.then(|| tdpp.position);
+                text_edit = Some(if indel.delete == source_range {
+                    self::completion_text_edit(line_index, insert_replace_support, indel.clone())
+                } else {
+                    assert!(source_range.end() == indel.delete.end());
+                    let range1 = TextRange::new(indel.delete.start(), source_range.start());
+                    let range2 = source_range;
+                    let indel1 = Indel::replace(range1, String::new());
+                    let indel2 = Indel::replace(range2, indel.insert.clone());
+                    additional_text_edits.push(self::text_edit(line_index, indel1));
+                    self::completion_text_edit(line_index, insert_replace_support, indel2)
+                })
             } else {
-                assert!(source_range.end() == indel.delete.end());
-                let range1 = TextRange::new(indel.delete.start(), source_range.start());
-                let range2 = source_range;
-                let indel1 = Indel::replace(range1, String::new());
-                let indel2 = Indel::replace(range2, indel.insert.clone());
-                additional_text_edits.push(self::text_edit(line_index, indel1));
-                self::completion_text_edit(line_index, insert_replace_support, indel2)
-            })
-        } else {
-            assert!(source_range.intersect(indel.delete).is_none());
-            let text_edit = self::text_edit(line_index, indel.clone());
-            additional_text_edits.push(text_edit);
+                assert!(source_range.intersect(indel.delete).is_none());
+                let text_edit = self::text_edit(line_index, indel.clone());
+                additional_text_edits.push(text_edit);
+            }
         }
-    }
-    let text_edit = text_edit.unwrap();
+        text_edit.unwrap()
+    };
 
     let mut lsp_item = lsp_types::CompletionItem {
         label: item.label().to_string(),
@@ -240,8 +262,57 @@ pub(crate) fn completion_item(
         ..Default::default()
     };
 
-    fn set_score(res: &mut lsp_types::CompletionItem, relevance: CompletionRelevance) {
-        if relevance.is_relevant() {
+    set_score(&mut lsp_item, max_relevance, item.relevance());
+
+    if item.deprecated() {
+        lsp_item.tags = Some(vec![lsp_types::CompletionItemTag::Deprecated])
+    }
+
+    if item.trigger_call_info() {
+        lsp_item.command = Some(command::trigger_parameter_hints());
+    }
+
+    if item.is_snippet() {
+        lsp_item.insert_text_format = Some(lsp_types::InsertTextFormat::Snippet);
+    }
+    if enable_imports_on_the_fly {
+        if let Some(import_edit) = item.import_to_add() {
+            let import_path = &import_edit.import.import_path;
+            if let Some(import_name) = import_path.segments().last() {
+                let data = lsp_ext::CompletionResolveData {
+                    position: tdpp.clone(),
+                    full_import_path: import_path.to_string(),
+                    imported_name: import_name.to_string(),
+                };
+                lsp_item.data = Some(to_value(data).unwrap());
+            }
+        }
+    }
+
+    if let Some((mutability, relevance)) = item.ref_match() {
+        let mut lsp_item_with_ref = lsp_item.clone();
+        set_score(&mut lsp_item_with_ref, max_relevance, relevance);
+        lsp_item_with_ref.label =
+            format!("&{}{}", mutability.as_keyword_for_ref(), lsp_item_with_ref.label);
+        if let Some(it) = &mut lsp_item_with_ref.text_edit {
+            let new_text = match it {
+                lsp_types::CompletionTextEdit::Edit(it) => &mut it.new_text,
+                lsp_types::CompletionTextEdit::InsertAndReplace(it) => &mut it.new_text,
+            };
+            *new_text = format!("&{}{}", mutability.as_keyword_for_ref(), new_text);
+        }
+
+        acc.push(lsp_item_with_ref);
+    };
+
+    acc.push(lsp_item);
+
+    fn set_score(
+        res: &mut lsp_types::CompletionItem,
+        max_relevance: u32,
+        relevance: CompletionRelevance,
+    ) {
+        if relevance.is_relevant() && relevance.score() == max_relevance {
             res.preselect = Some(true);
         }
         // The relevance needs to be inverted to come up with a sort score
@@ -253,39 +324,6 @@ pub(crate) fn completion_item(
         // tends to be since it is the opposite of the score.
         res.sort_text = Some(format!("{:08x}", sort_score));
     }
-
-    set_score(&mut lsp_item, item.relevance());
-
-    if item.deprecated() {
-        lsp_item.tags = Some(vec![lsp_types::CompletionItemTag::Deprecated])
-    }
-
-    if item.trigger_call_info() {
-        lsp_item.command = Some(command::trigger_parameter_hints());
-    }
-
-    let mut res = match item.ref_match() {
-        Some((mutability, relevance)) => {
-            let mut lsp_item_with_ref = lsp_item.clone();
-            set_score(&mut lsp_item_with_ref, relevance);
-            lsp_item_with_ref.label =
-                format!("&{}{}", mutability.as_keyword_for_ref(), lsp_item_with_ref.label);
-            if let Some(it) = &mut lsp_item_with_ref.text_edit {
-                let new_text = match it {
-                    lsp_types::CompletionTextEdit::Edit(it) => &mut it.new_text,
-                    lsp_types::CompletionTextEdit::InsertAndReplace(it) => &mut it.new_text,
-                };
-                *new_text = format!("&{}{}", mutability.as_keyword_for_ref(), new_text);
-            }
-            vec![lsp_item_with_ref, lsp_item]
-        }
-        None => vec![lsp_item],
-    };
-
-    for lsp_item in res.iter_mut() {
-        lsp_item.insert_text_format = Some(insert_text_format(item.insert_text_format()));
-    }
-    res
 }
 
 pub(crate) fn signature_help(
@@ -464,6 +502,7 @@ fn semantic_token_type_and_modifiers(
         },
         HlTag::Attribute => semantic_tokens::ATTRIBUTE,
         HlTag::BoolLiteral => semantic_tokens::BOOLEAN,
+        HlTag::BuiltinAttr => semantic_tokens::BUILTIN_ATTRIBUTE,
         HlTag::BuiltinType => semantic_tokens::BUILTIN_TYPE,
         HlTag::ByteLiteral | HlTag::NumericLiteral => lsp_types::SemanticTokenType::NUMBER,
         HlTag::CharLiteral => semantic_tokens::CHAR,
@@ -1174,7 +1213,9 @@ mod tests {
             encoding: OffsetEncoding::Utf16,
         };
         let (analysis, file_id) = Analysis::from_single_file(text);
-        let completions: Vec<(String, Option<String>)> = analysis
+
+        let file_position = ide_db::base_db::FilePosition { file_id, offset };
+        let mut items = analysis
             .completions(
                 &ide::CompletionConfig {
                     enable_postfix_completions: true,
@@ -1191,15 +1232,26 @@ mod tests {
                         skip_glob_imports: true,
                     },
                 },
-                ide_db::base_db::FilePosition { file_id, offset },
+                file_position,
             )
             .unwrap()
-            .unwrap()
-            .into_iter()
-            .filter(|c| c.label().ends_with("arg"))
-            .map(|c| completion_item(None, &line_index, c))
-            .flat_map(|comps| comps.into_iter().map(|c| (c.label, c.sort_text)))
-            .collect();
+            .unwrap();
+        items.retain(|c| c.label().ends_with("arg"));
+        let items = completion_items(
+            false,
+            false,
+            &line_index,
+            lsp_types::TextDocumentPositionParams {
+                text_document: lsp_types::TextDocumentIdentifier {
+                    uri: "file://main.rs".parse().unwrap(),
+                },
+                position: position(&line_index, file_position.offset),
+            },
+            items,
+        );
+        let items: Vec<(String, Option<String>)> =
+            items.into_iter().map(|c| (c.label, c.sort_text)).collect();
+
         expect_test::expect![[r#"
             [
                 (
@@ -1216,7 +1268,7 @@ mod tests {
                 ),
             ]
         "#]]
-        .assert_debug_eq(&completions);
+        .assert_debug_eq(&items);
     }
 
     #[test]
