@@ -8,6 +8,8 @@
 //! specific JSON shapes here -- there's little value in such tests, as we can't
 //! be sure without a real client anyway.
 
+mod sourcegen;
+mod tidy;
 mod testdir;
 mod support;
 
